/**
 * This class loads & stores the preferences for the webserver for use
 * by SatansHelpers (connection threads).
 *
 *
 * @version     $Id: PuzzleBox.java,v 1.1 2000/08/14 23:56:15 jst3290 Exp $
 *
 * @author      Joe Testa
 *
 *
 * Revisions:
 *              $Log: PuzzleBox.java,v $
 *              Revision 1.1  2000/08/14 23:56:15  jst3290
 *              Initial revision
 *
 */

package WebServer;
import java.io.*;
import java.util.*;
import java.lang.Integer;

import StartAllServers.Debug;
public class PuzzleBox {


    // The map to keep the preferences in.
    private Map preferencesMap;


    // These keys allow for the easy retrieval of data from the map.
    private final String HOSTNAME_KEY = "hostname";
    private final String WEBROOT_KEY = "webroot";
    private final String PORT_KEY = "port";
    private final String MAXCONNECTIONS_KEY = "maxconnections";
    private final String DIRECTORYBROWSING_KEY = "directorybrowsing";
    private final String ACCESSLOG_KEY = "accesslog";
    private final String ERRORLOG_KEY = "errorlog";
    private final String WEBMASTEREMAIL_KEY = "webmasteremail";
    private final String ERRORTEMPLATE_KEY = "errortemplate";
    private final String IPDENY_KEY = "ipdeny";
    private final String IPALLOW_KEY = "ipallow";
    private final String PRIORITY_KEY = "priority";
    
    

    /**
     * Main constructor.
     */

    public PuzzleBox( String fileName ) throws PuzzleBoxException {

        preferencesMap = new HashMap();

        StringTokenizer st;
        BufferedReader in;
        String aLine, theKey, theValue;

        try {

            // Reads the preferences file line by line, splits up the key/value
            //     pairs, and stores them in the Map.
            in = new BufferedReader( new FileReader( fileName ) );           
            while ( ( aLine = in.readLine() ) != null ) {

                st = new StringTokenizer( aLine, "=" );
                if ( st.hasMoreTokens() )  theKey = st.nextToken();
                else  theKey = "";
                if ( st.hasMoreTokens() )  theValue = st.nextToken();
                else  theValue = "";
                preferencesMap.put( theKey.trim(), theValue.trim() );

            }
            in.close();

        } catch ( FileNotFoundException e ) {
            throw( new PuzzleBoxException( "File not found:  " + fileName ) );
        } catch ( IOException e ) {
            throw( new PuzzleBoxException( "I/O error:  " + e.getMessage() ) );
        }



        int intValue = -1;

        try {

             //////////////////////////////////////////////////////////
            // Test the port value 

            intValue = getPort();

            if ( intValue > 65535 || intValue < 0 )
                throw( new Exception( "port field must be an integer " +
                                          "between 0 and 65535." ) );

            ///////////////////////////////////////////////////////////



             //////////////////////////////////////////////////////////
            // Test the maxconnections value 

            intValue = getMaxConnections();

            if ( intValue < 1 || intValue > 666 )
                throw( new Exception( "maxconnections field must be an " + 
                                          "integer between 1 and 666." ) );

            ///////////////////////////////////////////////////////////


             //////////////////////////////////////////////////////////
            // Test the priority value

            intValue = getPriority();
            if ( intValue == 777 )
                throw( new Exception( "priority field must be 'low', " +
                                          "'normal', or 'high'." ) );


            ///////////////////////////////////////////////////////////



            if ( isNull( WEBROOT_KEY ) || isNull( DIRECTORYBROWSING_KEY ) ||
                 isNull( ACCESSLOG_KEY ) || isNull( ERRORLOG_KEY ) ||
                 isNull( ERRORTEMPLATE_KEY ) || isNull( PRIORITY_KEY )) {
                     throw( new Exception( "blank setting in one or more " +
                                               "required fields." ) );
            }



            if ( !isNull( IPDENY_KEY ) && !isNull( IPALLOW_KEY ) )
                throw( new Exception( "illegal to specify ipdeny and " + 
                                          "ipallow fields simultaneously." ) );



            if ( isNull( HOSTNAME_KEY ) )
                preferencesMap.put( HOSTNAME_KEY, "localhost" );



        } catch ( Exception e ) {
            throw( new PuzzleBoxException( 
                       "Invalid, incomplete, or corrupt preferences file:  " +
                       fileName + ".  Reason: " + e.getMessage() ) );
        }


    }



    /**
     * Returns whether or not directory browsing is allowed.
     *
     * @returns     'true' if directory browsing is allowed; 'false' if
     *                  otherwise.
     */

    public boolean getDirectoryBrowsing() {

        String theValue =
                     ( String )( preferencesMap.get( DIRECTORYBROWSING_KEY ) );
        return( theValue.equals( "on" ) );

    }



    /**
     * Returns the hostname of this server.
     *
     * @return       the hostname of this server.
     */

    public String getHostName() {
        return ( String )( preferencesMap.get( HOSTNAME_KEY ) );
    }



    /**
     * Returns the number of maximum connections that this server can
     *     handle at one time.
     *
     * @return      the number of maximum connections that this server can
     *                  handle at one time.
     */

    public int getMaxConnections() {


        String stringValue = "";
        int intValue = -1;

        try {

            stringValue =
                        ( String )( preferencesMap.get( MAXCONNECTIONS_KEY ) );
            intValue = Integer.parseInt( stringValue );

        } catch ( NumberFormatException e ) {}

        return intValue;


    }



    /**
     * Returns the port number this server is set to listen on.
     *
     * @return           the port number this server is set to listen on.
     */

    public int getPort() {


        String stringValue = "";
        int intValue = -1;

        try {

            stringValue = ( String )( preferencesMap.get( PORT_KEY ) );
            intValue = Integer.parseInt( stringValue );

        } catch ( NumberFormatException e ) {}

        return intValue;


    }



    /**
     * Returns the priority that the server's threads are set to run at.
     *
     * @return        priority that the server's threads are set to run at.
     */

    public int getPriority() {


        String stringPriority = 
                              ( String )( preferencesMap.get( PRIORITY_KEY ) );
        int threadPriority = 777;

        if ( stringPriority.equals( "low" ) )
            threadPriority = Thread.MIN_PRIORITY;
        else if ( stringPriority.equals( "normal" ) )
            threadPriority = Thread.NORM_PRIORITY;
        else if ( stringPriority.equals( "high" ) )
            threadPriority = Thread.MAX_PRIORITY;

        return threadPriority;


    }



    /**
     * Returns the webmaster's email address.
     *
     * @return      the webmaster's email address.
     */

    public String getWebmasterEmail() {
        return ( String )( preferencesMap.get( WEBMASTEREMAIL_KEY ) );
    }



    /**
     * Returns the root directory that this server is to make available.
     *
     * @return       the root directory that this server is to make available.
     */

    public String getWebRoot() {
        return ( String )( preferencesMap.get( WEBROOT_KEY ) );
    }



    /**
     * Returns the path to the access log.
     *
     * @return      the path to the access log.
     */

    public String getAccessLog() {
        return ( String )( preferencesMap.get( ACCESSLOG_KEY ) );
    }



    /**
     * Returns the path to the error log.
     *
     * @return      the path to the error log.
     */

    public String getErrorLog() {
        return ( String )( preferencesMap.get( ERRORLOG_KEY ) );
    }



    /**
     * Returns the path to the error template.
     *
     * @return      the path to the error template.
     */

    public String getErrorTemplate() {
        return ( String )( preferencesMap.get( ERRORTEMPLATE_KEY ) );
    }



    /**
     * Returns a (linked) list of IP addresses to allow.
     *
     * @return      a list of IP addresses to allow.
     */

    public List getIPAllow() {


        String ipStrings;
        StringTokenizer st;
        LinkedList ipList;

        ipList = new LinkedList();
        ipStrings = ( String )( preferencesMap.get( IPALLOW_KEY ) );

        st = new StringTokenizer( ipStrings, ";" );
        while( st.hasMoreTokens() )  ipList.addLast( st.nextToken().trim() );

        return ipList;


    }



    /**
     * Returns a list of IP addresses to deny.
     *
     * @return      a list of IP addresses to deny.
     */

    public List getIPDeny() {


        String ipStrings;
        StringTokenizer st;
        LinkedList ipList;

        ipList = new LinkedList();
        ipStrings = ( String )( preferencesMap.get( IPDENY_KEY ) );

        st = new StringTokenizer( ipStrings, ";" );
        while( st.hasMoreTokens() )  ipList.addLast( st.nextToken().trim() );

        return ipList;


    }



    /**
     * Returns the MIME type for the specified suffix.
     *
     * @return      the MIME type for the specified suffix.
     */

    public String getMIMEType( String suffix ) {
        return ( String )( preferencesMap.get( suffix ) );
    }



    /**
     * Returns a string representation of this PuzzleBox.
     *
     * @return      a string representation of this PuzzleBox.
     */

    public String toString() {



        int intPriority = -1;
        String stringPriority = "";

        intPriority = getPriority();
        if ( intPriority == Thread.MIN_PRIORITY )
            stringPriority = "low";
        else if ( intPriority == Thread.NORM_PRIORITY )
            stringPriority = "normal";
        else if ( intPriority == Thread.MAX_PRIORITY )
            stringPriority = "high";


        String stuff = "";

        stuff += DIRECTORYBROWSING_KEY + ":  " + getDirectoryBrowsing() + "\n";
        stuff += HOSTNAME_KEY + ":           " + getHostName() + "\n";
        stuff += MAXCONNECTIONS_KEY + ":     " + getMaxConnections() + "\n";
        stuff += PORT_KEY + ":               " + getPort() + "\n";
        stuff += PRIORITY_KEY + ":           " + stringPriority + "\n";
        stuff += WEBROOT_KEY + "             " + getWebRoot() + "\n"; 
        stuff += ACCESSLOG_KEY + ":          " + getAccessLog() + "\n";
        stuff += ERRORLOG_KEY + ":           " + getErrorLog() + "\n";
        stuff += WEBMASTEREMAIL_KEY + ":     " + getWebmasterEmail() + "\n";
        stuff += ERRORTEMPLATE_KEY + ":      " + getErrorTemplate() + "\n";
        stuff += IPDENY_KEY + ":             " + getIPDeny() + "\n";
        stuff += IPALLOW_KEY + ":            " + getIPAllow() + "\n";

        return stuff;


    }



    /**
     * Tests if the specified string key is defined in the internal Map.
     *
     * @param      theKey     the string key to test.
     * @return                'true' if the key is defined, 'false' if 
     *                            otherwise.
     */

    private boolean isNull( String theKey ) {

            String stringValue = ( String )( preferencesMap.get( theKey ) );
            return( stringValue.equals( "" ) );

    }



}  // PuzzleBox
