/**
 * Main file for hellbent java webserver.
 *
 *
 * @version      $Id: SimpleServer.java,v 1.1 2000/08/14 23:57:06 jst3290 Exp $
 *
 * @author       Joe Testa
 *
 *
 * Revisions:
 *               $Log: SimpleServer.java,v $
 *               Revision 1.1  2000/08/14 23:57:06  jst3290
 *               Initial revision
 *
 */

package WebServer;
import java.util.*;
import java.net.*;
import java.io.*;

import StartAllServers.Debug;
public class SimpleServer extends Thread {



    // Name and version of this webserver.
    public static final String PROGRAM_NAME =
                                            "[ SimpleServer java webserver v0.1 ]";
    // Path to preferences file.
    public static final String PREFS_PATH = "./WebServer/SimpleServer.prefs";

    // Prompt for console.
    public static final String CONSOLE_PROMPT = "SimpleServer> ";

    // Timeout for 'accept()' method.
    public static final int LISTEN_TIMEOUT = 5000;

    // Array of pictures packaged with this version.
    public static final String[] pictureArray = { "hellbent.jpg",
                                                  "poweredby.jpg",
                                                  "recommended.jpg" };

    // If 'true', extra info is printed to the console from time to time.
    public static boolean verboseFlag = false;



    private static boolean continueFlag;  // 'false' when server should quit.
    private static PuzzleBox puzzleBox;   // PuzzleBox to manage preferences.
    private static PrintWriter out;       // PrintWriter to write logs.
    private static String accessLogPath;  // Path to access log.
    private static String errorLogPath;   // Path to error log.
    private static long startTime;        // Time that the server started.
    private static int numThreads;        // Number of ServerHelpers.
    private static int totalConnections;  // Total number of connections made
                                          //     since server was started.
    private static LinkedList threadList; // A list of threads (ServerHelpers).
    private static int maxConnections;    // Max number of connections allowed.
    private static int threadPriority;    // Priority to set each thread.


    private ServerSocket serverSocket;    // Server socket to listen with.
    private Socket incommingConnection;   // Normal socket to pass to
                                          //     ServerHelpers.
    private ServerHelper serverHelper;    // Variable to spawn ServerHelpers.




    public SimpleServer() {


        int portNumber = puzzleBox.getPort();

        // Create a new server socket.
        try {

            serverSocket = new ServerSocket( portNumber, maxConnections );
            serverSocket.setSoTimeout( LISTEN_TIMEOUT );

        } catch( SocketException e ) {
            System.err.print( "Could not create server socket on port " );
            System.err.print( portNumber + " with backlog queue of " );
            System.err.println( maxConnections + "." );
            continueFlag = false;
        } catch( IOException e ) {
            System.err.println( "I/O error:  " + e.getMessage() );
            continueFlag = false;
        }
    

    } 



    /**
     * The run method.
     */

    public void run() {


        // While server has not been shut down...
        while( continueFlag ) {
            try {


                // Update the list of threads and check if any more connections
                //     should be accepted.
                if ( checkThreads() ) {


                    // Wait for incomming connection and give 
                    //     'incommingConnection' a reference to the new socket.
                    //     'accept()' will block only for LISTEN_TIMEOUT
                    //     milliseconds before falling out of 'try' block and
                    //     possibly repeating.

                    incommingConnection = serverSocket.accept();


                    if ( SimpleServer.verboseFlag ) {
if (Debug.DEBUG_WebServer == true) { 
 	 System.out.print( "    - Connection established; " ); 
 	}
if (Debug.DEBUG_WebServer == true) { 
 	 System.out.print( "spawning thread... " ); 
 	}
                    }


                    // Make a new ServerHelper and give it a reference to the
                    //     socket.  Then make the thread a daemon, set its
                    //     priority, and start it.

                    serverHelper = new ServerHelper( incommingConnection );
                    serverHelper.setDaemon( true );
                    serverHelper.setPriority( threadPriority );
                    serverHelper.start();

                    // Add this thread to the thread list.
                    addThread( serverHelper );      

                    if ( SimpleServer.verboseFlag )
if (Debug.DEBUG_WebServer == true) { 
 	 System.out.println( "done." ); 
 	}


                } else        // If 'maxConnections' has been reached, yield
                    yield();  //     until a spot opens.


            }

            // This exception is thrown when LISTEN_TIMEOUT milliseconds have
            //     elapsed.  In this case, nothing should be done because
            //     execution will fall out of the 'try' block so that
            //     'continueFlag' is checked by the loop.
            catch ( InterruptedIOException e ) { }  
            catch ( IOException e ) {  continueFlag = false;  }
        }


    }



    /**
     * Adds a thread to the internal thread list and increments the thread
     *     counter.
     *
     * @param    aThread    the thread to add
     */

    private static void addThread( Thread aThread ) {


        // Add the thread to the LinkedList.
        threadList.addLast( aThread );  
        numThreads++;              // Increase the current number of threads.
        totalConnections++;        // Increase the total number of serviced
                                   //     connections.

    }



    /**
     * Purges the internal thread list of dead threads, updates the thread
     *     counter, and checks if any more threads should be allowed to
     *     spawn based on the maxConnections variable.
     *
     * @return              'true' if maxConnections has not been exeeded;
     *                          'false' if it has.
     */

    private static synchronized boolean checkThreads() {

        ServerHelper aThread;
        for ( int i = 0; i < threadList.size(); i++ ) {


            // Check if each thread is alive.  If not, remove it from the list,
            //     kill its connection (just in case), and decrement the number
            //     of running threads.

            aThread = ( ServerHelper )threadList.get( i );
            if ( !aThread.isAlive() ) {
                 threadList.remove( i );
                 aThread.killConnection();
                 numThreads--;
            }

        }

        // Return true if more connections are allowed, false if otherwise.
        return ( numThreads < maxConnections );


    }



    /**
     * Returns a random picture filename.
     *
     * @return     the filename of a random picture.
     */

    public static String getRandomPicture() {

        return pictureArray[ ( int )( Math.random() * pictureArray.length ) ];

    }



    /**
     * Loads the preferences using a PuzzleBox
     */

    private static void loadPrefs() {

        try {

            // Make a PuzzleBox and give it the path to the preferences file.
            puzzleBox = new PuzzleBox( PREFS_PATH );
            ServerHelper.puzzleBox = puzzleBox;
if (Debug.DEBUG_WebServer == true) { 
 	 System.out.println( puzzleBox.toString() ); 
 	}


            numThreads = 0;
            totalConnections = 0;
            continueFlag = true;


            maxConnections = puzzleBox.getMaxConnections();
            threadList = new LinkedList();

            accessLogPath = puzzleBox.getAccessLog();
            errorLogPath = puzzleBox.getErrorLog();

            threadPriority = puzzleBox.getPriority();


        } catch ( PuzzleBoxException e ) {
            System.err.println( e.getMessage() );
            System.exit( 777 );
        }
        
    }



    /**
     * Records a connection to the log file.
     *
     * @param     hostAddress     the host address of the client.
     * @param     date            the time and date.
     * @param     input           the client's request.
     */

    public static synchronized void logAccess( String hostAddress,
                                               String date,
                                               String input )  {


        // Use a PrintWriter to print an entry to the access log file.
        try {

            out = new PrintWriter( new BufferedWriter(
                                   new FileWriter( accessLogPath, true ) ) );
            out.println( hostAddress + " - [" + date + "] - \"" + input +
                                                                      "\"" );
            out.flush();
            out.close();

        } catch( IOException e ) {
            System.err.println( "I/O error:  " + e.getMessage() );
        }        

    }



    /**
     * Records an error to the error file
     *
     * @param     hostAddress    the host address of the client.
     * @param     date           the time and date.
     * @param     error          the error which occured.
     */

    public static synchronized void logError( String hostAddress,
                                              String date,
                                              int errorCode,
                                              String errorText,
                                              String firstLine ) {


        // Use a PrintWriter to print an entry to the error log file.

        try {

            out = new PrintWriter( new BufferedWriter(
                                   new FileWriter( errorLogPath, true ) ) );
            out.println( hostAddress + " - [" + date + "] - " + errorCode +
                             ": " + errorText + "  \"" + firstLine + "\"" );
            out.flush();
            out.close();

        } catch( IOException e ) {
            System.err.println( "I/O error:  " + e.getMessage() );
        }        


    }



    /**
     * Reads and processes commands from the console.
     */

    private static void processConsoleCommands() {


        BufferedReader in = new BufferedReader(
                            new InputStreamReader( System.in ) );
        String aLine = null;

        try {


            do {

if (Debug.DEBUG_WebServer == true) { 
 	 System.out.print( CONSOLE_PROMPT ); 
 	}
                aLine = ( in.readLine() ).toLowerCase();



                if ( aLine.equals( "help" ) || aLine.equals( "?" ) ) {

                    // List the available commands.
if (Debug.DEBUG_WebServer == true) { 
 	 System.out.println(); 
 	}
                    System.out.println( "    help | ?                        " +
                                                   " - prints this help info" );
                    System.out.println( "    view [ access | error | prefs ] " +
                                        " - prints out the access log, error" );
                    System.out.println( "                                    " +
                                        "        log, or current preferences" );
                    System.out.println( "    stat                            " +
                                               " - prints status information" );
                    System.out.println( "    verbose                         " +
                                                 " - toggles verbose setting" );
if (Debug.DEBUG_WebServer == true) { 
 	 System.out.println(); 
 	}
                    System.out.println( "    exit | quit                     " +
                                                  " - terminates the program" );
if (Debug.DEBUG_WebServer == true) { 
 	 System.out.println(); 
 	}

                }



                if ( aLine.equals( "stat" ) ) {


                    // Print some stats to the user.
                    long currentTime = System.currentTimeMillis() - startTime;

                    int seconds = ( int )( currentTime / 1000 ) % 60;
                    int minutes = ( int )( currentTime / 60000 ) % 60;
                    int hours = ( int )( currentTime / 3600000 );
                    //int days = ( int )( currentTime / 86400000 );

                    // Update list of connection threads.
                    checkThreads();

                    
                    String stringPriority = "";
                    if ( threadPriority == Thread.MIN_PRIORITY )
                        stringPriority = "low";
                    else if ( threadPriority == Thread.NORM_PRIORITY )
                        stringPriority = "normal";
                    else if ( threadPriority == Thread.MAX_PRIORITY )
                        stringPriority = "high";



if (Debug.DEBUG_WebServer == true) { 
 	 System.out.println(); 
 	}
                    System.out.println( "    - Number of threads running (at " +
                                               stringPriority + " priority):" );
                    System.out.println( "        " + numThreads +
                                           " (connection) + 2 (core program).");
                    System.out.print( "    - Server has been running for " +
                                                           hours + " hours, " );
                    System.out.println( minutes + " minutes, " + seconds +
                                                              " seconds and " );
                    System.out.println( "        serviced " + totalConnections +
                                                              " connections." );
                    System.out.println( "    - Verbose: " +
                                               ( verboseFlag ? "on" : "off" ) );
if (Debug.DEBUG_WebServer == true) { 
 	 System.out.println(); 
 	}

       
                }



                if ( aLine.equals( "verbose" ) ) {
                    verboseFlag = !verboseFlag;
if (Debug.DEBUG_WebServer == true) { 
 	 System.out.println(); 
 	}
                    System.out.println( "    - Verbose is now " +
                                             ( verboseFlag ? "on." : "off." ) );
if (Debug.DEBUG_WebServer == true) { 
 	 System.out.println(); 
 	}
                }



                if ( aLine.equals( "view access" ) ) {


                    // Dump the access log to the console (20 lines at a time).
                    BufferedReader theFile = null;
                    String fileLine = null;
                    int i = 0;

                    try {

                      theFile = new BufferedReader(
                                new FileReader( accessLogPath ) );

                      while ( ( fileLine = theFile.readLine() ) != null ) {
if (Debug.DEBUG_WebServer == true) { 
 	 System.out.println( fileLine ); 
 	}  i++;
                          if ( i == 20 ) {
                              System.out.println( "    - Press enter to view " +
                                                            "next 20 entries" );
                              in.readLine();
                              i = 0;
                          }
                      }

                      theFile.close();

                    } catch ( FileNotFoundException e ) {
                          System.err.println( e.getMessage() );
                    }


                }



                if ( aLine.equals( "view error" ) ) {


                    // Dump the error log to the console (20 lines at a time).
                    BufferedReader theFile = null;
                    String fileLine = null;
                    int i = 0;

                    try {

                      theFile = new BufferedReader(
                                new FileReader( errorLogPath ) );
                      while ( ( fileLine = theFile.readLine() ) != null ) {
if (Debug.DEBUG_WebServer == true) { 
 	 System.out.println( fileLine ); 
 	}  i++;
                          if ( i == 20 ) {
                              System.out.println( "    - Press enter to view " +
                                                            "next 20 entries" );
                              in.readLine();
                              i = 0;
                          }
                      }

                      theFile.close();

                    } catch ( FileNotFoundException e ) {
                          System.err.println( e.getMessage() );
                    }



                }


                if ( aLine.equals( "view prefs" ) ) {

                    // Print out the preferences.
if (Debug.DEBUG_WebServer == true) { 
 	 System.out.println(); 
 	}
if (Debug.DEBUG_WebServer == true) { 
 	 System.out.println( puzzleBox.toString() ); 
 	}

                }



            } while ( !aLine.equals( "exit" ) && !aLine.equals( "quit" ) );


            in.close();
            shutDownServer();


        } catch( IOException e ) {
            System.err.println( "I/O error:  " + e.getMessage() );
        }


    }



    /**
     * Shuts down the server.
     */

    private static void shutDownServer() {
if (Debug.DEBUG_WebServer == true) { 
 	 System.out.println(); 
 	}
if (Debug.DEBUG_WebServer == true) { 
 	 System.out.println( "Shutting down server..." ); 
 	}
        continueFlag = false;
    }

    
    public static void startWebServer() {
	
	//  if ( args.length != 0 ) {
        //    System.out.println(
	//                     "Error:  command line arguments are not used." );
        //    System.exit( 777 );
        //}


if (Debug.DEBUG_WebServer == true) { 
 	 System.out.println(); 
 	}
if (Debug.DEBUG_WebServer == true) { 
 	 System.out.println( PROGRAM_NAME ); 
 	}
if (Debug.DEBUG_WebServer == true) { 
 	 System.out.println(); 
 	}
if (Debug.DEBUG_WebServer == true) { 
 	 System.out.println(); 
 	}
if (Debug.DEBUG_WebServer == true) { 
 	 System.out.println( "\tLoading preferences..." ); 
 	}
if (Debug.DEBUG_WebServer == true) { 
 	 System.out.println(); 
 	}
        loadPrefs();
        

if (Debug.DEBUG_WebServer == true) { 
 	 System.out.print( "\tStarting server... " ); 
 	}

        Thread mySimpleServer = new SimpleServer();
        mySimpleServer.setDaemon( true );
        mySimpleServer.setPriority( threadPriority );
        mySimpleServer.start();

        startTime = System.currentTimeMillis();

if (Debug.DEBUG_WebServer == true) { 
 	 System.out.println( "done." ); 
 	}


if (Debug.DEBUG_WebServer == true) { 
 	 System.out.println(); 
 	}
if (Debug.DEBUG_WebServer == true) { 
 	 System.out.println(); 
 	}
        processConsoleCommands();
        System.exit( 666 );


    }



    /**
     * The main method.
     *
     * @param     args[]     command line arguements (not used).
     */

    public static void main( String args[] ) {


        if ( args.length != 0 ) {
            System.out.println(
                              "Error:  command line arguments are not used." );
            System.exit( 777 );
        }


if (Debug.DEBUG_WebServer == true) { 
 	 System.out.println(); 
 	}
if (Debug.DEBUG_WebServer == true) { 
 	 System.out.println( PROGRAM_NAME ); 
 	}
if (Debug.DEBUG_WebServer == true) { 
 	 System.out.println(); 
 	}
if (Debug.DEBUG_WebServer == true) { 
 	 System.out.println(); 
 	}
if (Debug.DEBUG_WebServer == true) { 
 	 System.out.println( "\tLoading preferences..." ); 
 	}
if (Debug.DEBUG_WebServer == true) { 
 	 System.out.println(); 
 	}
        loadPrefs();
        

if (Debug.DEBUG_WebServer == true) { 
 	 System.out.print( "\tStarting server... " ); 
 	}

        Thread mySimpleServer = new SimpleServer();
        mySimpleServer.setDaemon( true );
        mySimpleServer.setPriority( threadPriority );
        mySimpleServer.start();

        startTime = System.currentTimeMillis();

if (Debug.DEBUG_WebServer == true) { 
 	 System.out.println( "done." ); 
 	}


if (Debug.DEBUG_WebServer == true) { 
 	 System.out.println(); 
 	}
if (Debug.DEBUG_WebServer == true) { 
 	 System.out.println(); 
 	}
        processConsoleCommands();
        System.exit( 666 );


    }



}  // SimpleServer
